/*---------------------------------------------------------------------

mynoise.cc

This file is part of CytoPacq

Copyright (C) 2007,2008 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Author: David Svoboda

Description: Generation of various types of noise.

-------------------------------------------------------------------------*/

#include "mynoise.h"

#ifdef __GNUG__
	#include <values.h>
#endif

#ifdef _MSC_VER
	#include <time.h>
#endif

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

using namespace i3d;
using namespace std;

/****************************************************************************/
template <class ELECTRONS> 
void ApplyPoissonNoise(Image3d<ELECTRONS> &img)
{
	ELECTRONS el_amount,out_intensity;

	static gsl_rng *randState=NULL;

	//do we need to init random generator?
	if (randState == NULL) {
		//yes, we do:
		//create instance of the generator and seed it
		randState = gsl_rng_alloc(gsl_rng_default);
		unsigned long s=-1 * (int) time(NULL);
		gsl_rng_set(randState,s);
	}

	size_t image_size=img.GetImageSize();

	for(size_t l=0; l<image_size; l++)
	{
		el_amount = img.GetVoxel(l);

		double noiseMean = sqrt((double)el_amount);
		double noiseVar = noiseMean;

		out_intensity = static_cast<ELECTRONS>(
									ceilf((double) el_amount 
									+ gsl_ran_poisson(randState, noiseMean) 
									- noiseVar));

		img.SetVoxel(l,out_intensity);
	}
}

/****************************************************************************/

template <class ELECTRONS> 
void ApplyReadoutNoise(Image3d<ELECTRONS> &img, float sigma)
{
	ELECTRONS in_intensity,out_intensity;

	static gsl_rng *randState=NULL;

	//do we need to init random generator?
	if (randState == NULL) 
	{
		//yes, we do:
		//create instance of the generator and seed it
		randState = gsl_rng_alloc(gsl_rng_default);
		unsigned long s=-1 * (int) time(NULL);
		gsl_rng_set(randState,s);
	}

	size_t image_size=img.GetImageSize();

	for(size_t l=0; l<image_size; l++){
		in_intensity = img.GetVoxel(l);

		out_intensity = (ELECTRONS)
			round(gsl_ran_gaussian(randState, sigma) + (double) in_intensity);

		img.SetVoxel(l, out_intensity);
	}
}

//----------------------------------------------------------------------------
template <class ELECTRONS> 
void SetAmountOfElectrons(Image3d<ELECTRONS> &img, 
								  size_t x, 
								  size_t y, 
								  ELECTRONS amount)
{
	for (size_t z=0; z<img.GetSizeZ(); z++)
	{
		img.SetVoxel(x, y, z, amount);
	}
}

//----------------------------------------------------------------------
template <class ELECTRONS>
void AddDarkCurrent(Image3d<ELECTRONS> &image,  float el_added, float ac_time)
{
	// el_added ... amount of electrons per pixel per second
	// ac_time ... acquisition time (in miliseconds)
	 
	static gsl_rng *randState=NULL;

	//do we need to init random generator?
	if (randState == NULL) 
	{
		//yes, we do:
		//create instance of the generator and seed it
		randState = gsl_rng_alloc(gsl_rng_default);
		unsigned long s=-1 * (int) time(NULL);
		gsl_rng_set(randState,s);
	}
	
	double dark_current_mean = el_added * ac_time * 0.001; 

	for (size_t i=0; i<image.GetImageSize(); i++)
	{		 
		 ELECTRONS real_dark_current = gsl_ran_poisson(randState, dark_current_mean);
		 image.GetVoxelData()[i] += real_dark_current;
	}
}

//----------------------------------------------------------------------------
/*explicit instantiations */
//----------------------------------------------------------------------------

template  void SetAmountOfElectrons(Image3d<GRAY16> &, size_t , size_t, GRAY16);
template  void SetAmountOfElectrons(Image3d<GRAY8> &, size_t , size_t, GRAY8);
template  void SetAmountOfElectrons(Image3d<size_t> &, size_t , size_t, size_t);
//template  void ApplyPoissonNoise(Image3d<GRAY8> &img);
//template  void ApplyPoissonNoise(Image3d<GRAY16> &img);
template  void ApplyPoissonNoise(Image3d<size_t> &img);
//template  void ApplyReadoutNoise(Image3d<GRAY8> &img, float sigma);
//template  void ApplyReadoutNoise(Image3d<GRAY16> &img, float sigma);
template  void ApplyReadoutNoise(Image3d<size_t> &img, float sigma);
template  void AddDarkCurrent(Image3d<size_t> &,  float, float);
//----------------------------------------------------------------------------
