/*---------------------------------------------------------------------

plug.cc

This file is part of CytoPacq

Copyright (C) 2007,2008 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Author: David Svoboda

Description: A module responsible for plugins loading.

-------------------------------------------------------------------------*/

#include <iostream>
#include <string.h>
#include <stdlib.h>

#ifdef _WIN32
	#include <windows.h>
	#define BUFSIZE MAX_PATH
#else
	#include <unistd.h>
	#include <dirent.h>
	#include <fnmatch.h>
	#define BUFSIZE PATH_MAX
#endif

#include "sim_macros.h"

#include "plug.h"

using namespace std;

/*****************************************************************************/

T_HANDLE FindAndOpenPlugin(const char* identifier, 
									const std::vector<std::string> &fnames)
{
	 typedef const char* (*t_string)(void);

	 // Inspect all available plugins
	 for (size_t k=0; k<fnames.size(); k++)
	 {
		  // try to open the plugin file
		  echo("opening plugin: " << fnames[k]);

		  T_HANDLE plugin_handle = OpenPlugin(fnames[k].c_str());

		  if (plugin_handle == NULL)
				throw "bad plugin";

		  // read the necessary symbols from the plugin file
		  echo("reading plugin symbols");

		  t_string plugin_id = 
			 (t_string) GetSymbolFromPlugin(plugin_handle, "get_id");

		  if (plugin_id == NULL)
				continue;

		  echo("identifier: " << plugin_id());

		  // is this plugin the one we are looking for?
		  if (strcmp(plugin_id(), identifier) == 0)
		  {

				echo("plugin with id >>>>" << identifier << "<<<< found");
				return plugin_handle;
		  }

		  echo("identifiers do not match");

		  // next plugin
	 }

	return NULL;
}

/*****************************************************************************/

T_HANDLE OpenPlugin(const char* fn)
{
	 T_HANDLE fhandle;

#ifdef _WIN32
		  fhandle = LoadLibrary(fn);
		  if (fhandle == NULL)
		  {
			  echo("Error: " << GetLastError());
				return NULL;
		  }
#else
		  fhandle = dlopen(fn, RTLD_NOW);

		  if (fhandle == NULL) // ignore bad file
		  {
				echo(dlerror());
				return NULL;
		  }
#endif

		  return fhandle;
}

/*****************************************************************************/

void *GetSymbolFromPlugin(T_HANDLE fhandle, const char *symbol_name)
{
	 void *ptr;

#ifdef _WIN32
	 ptr = GetProcAddress(fhandle, symbol_name);
	
	 if (ptr == NULL)
	 {
		 echo("Error: " << GetLastError());
		  return NULL;
	 }
#else 
	 ptr = dlsym(fhandle, symbol_name);

	 if (ptr == NULL)
	 {
		  echo(dlerror());
		  return NULL;
	 }
#endif

	 return ptr;
}

/*****************************************************************************/
bool VerifyPlugin(std::string &fname)
{
	 T_HANDLE fhandle;

	 // check opening the file
	 if ((fhandle = OpenPlugin(fname.c_str())) == NULL)
		  return false;

	 // check the file conent for correctness
	 if (GetSymbolFromPlugin(fhandle, "create") == NULL)
		  return false;

	 // everything is ok -> close the file a return true value
	 ReleasePlugin(fhandle);
	 return true;
}

/*****************************************************************************/

void ReleasePlugin(T_HANDLE fhandle)
{
#ifdef _WIN32
	 FreeLibrary(fhandle);
#else
	 dlclose(fhandle);
#endif
}

/*****************************************************************************/

#ifndef _WIN32
int is_a_plugin (const struct dirent *item)
{
	 if (!fnmatch("lib*so", item->d_name, 0))
		  return 1;

	return 0;
}
#endif

/*****************************************************************************/

int SearchForPlugins(std::vector<std::string> &plugin_list,
							const char *directory)
{
	 std::string abs_path, full_name;

#ifdef _WIN32 // WINDOWS ONLY
	 if (isupper(directory[0]) && (directory[1] == ':'))
	 {
		  abs_path = directory;
	 }
	 else
	 {
		 // set up the full name of the directory
		 char buf[BUFSIZE];
		 GetCurrentDirectory(BUFSIZE, buf);
		 abs_path = string(buf) + string("\\") + directory;
	 }

	 std::string pattern = abs_path + "\\*dll";

	 WIN32_FIND_DATA FindFileData;
	 HANDLE hFind = FindFirstFile(pattern.c_str(), &FindFileData);

	 if (hFind == INVALID_HANDLE_VALUE)
	 {
		  echo("Warning: reading directory failed");
		  return 0;
	 }

	 do 
	 {
		 full_name = abs_path + "\\" + FindFileData.cFileName;

		 if (VerifyPlugin(full_name))
		 {
			 plugin_list.push_back(full_name);
		 }
	 } while (FindNextFile(hFind, &FindFileData) != 0);

	 FindClose(hFind);

#else // LINUX ONLY
	 if (directory[0] == '/')
	 {
		  abs_path = directory;
	 }
	 else
	 {
		 // set up the full name of the directory
		 char buf[BUFSIZE];
		 abs_path = getcwd(buf, BUFSIZE);
		 abs_path += string("/") + directory;
	 }
	 // echo("selected directory: " << abs_path.c_str());

	 struct dirent **namelist = NULL;
    int count = scandir(abs_path.c_str(), 
								&namelist, 
								is_a_plugin, 
								alphasort);

	 if (count < 0)
	 {
		  echo("Warning: reading directory failed");
		  return 0;
	 }

	 for (int k=0; k<count; k++)
	 {
		  string full_name = abs_path + string("/") + namelist[k]->d_name;
		  if (VerifyPlugin(full_name))
		  {
				plugin_list.push_back(full_name);
		  }
		  free(namelist[k]);
	 }

	 free(namelist);
#endif

	 return plugin_list.size();
}


/*****************************************************************************/
