/*---------------------------------------------------------------------

hl60.cc

This file is part of CytoPacq

Copyright (C) 2007-2013 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Author: David Svoboda

Description: A plugin creating the digital phantom of HL60 nucleus.

-------------------------------------------------------------------------*/

#include <i3d/image3d.h>
#include <i3d/threshold.h>
#include <i3d/draw.h>

#include "sim_macros.h"

#include "src/add_shape.h"
#include "src/deformation.h"
#include "src/texture.h"
#include "src/myrandom.h"
#include "src/collisions.h"

#include "hl60.h"

using namespace i3d;

//------------------------------------------------------------------------
extern "C" DLL_EXPORT const char *get_id()
{
	 return "HL60 nucleus";
}

//------------------------------------------------------------------------
void phantom_hl60::read_params(const t_item_seq &phantom_properties,
										 const t_item_seq &common_properties)
{
	 phantom_with_position::read_params(phantom_properties,
													common_properties);

#ifdef DEBUG
    // read the object parameters from the input
	 radius = phantom_properties["semi-axes"];
	 r_variance = phantom_properties["semi-axes variance"];
	 rot_freedom = phantom_properties["object rotation freedom"];
    ishape = phantom_properties["shape irregularity"];
	 ishape_var = phantom_properties["shape irregularity variance"];
	 num_of_nucleoli = phantom_properties["number of nucleoli"];
	 size_of_nucleoli = phantom_properties["size of nucleoli"];
	 blob_size_ref = phantom_properties["structure blob size"];
	 blob_size_var = phantom_properties["structure blob size variance"];
	 smoothness_ref = phantom_properties["structure smoothness"];
	 smoothness_var = phantom_properties["structure smoothness variance"];
	 contrast = phantom_properties["structure contrast"];
	 condensation_var = phantom_properties["condensation variance"];
#else
	 // use the predefined values
	 radius.x = 5.0f;
	 radius.y = 5.0f;
	 radius.z = 4.0f;
	 r_variance = 28;
	 rot_freedom = 50;
	 ishape = 2.0f;
	 ishape_var = 15.0f;
	 num_of_nucleoli = 2;
	 size_of_nucleoli = 33;
	 blob_size_ref = 1.2f;
	 blob_size_var = 40.0f;
	 smoothness_ref = 8.0f;
	 smoothness_var = 10.0f;
	 contrast = 2.0f;
	 condensation_var = 30.0f;
#endif

}

//------------------------------------------------------------------------

bool phantom_hl60::draw(i3d::Image3d<i3d::GRAY8>& img_phantoms, 
								i3d::Image3d<LABEL_TYPE>& img_labels)
{
	 if (!_create_shape(img_labels))
		  return false; 
	 
	 _create_texture(img_labels, img_phantoms); 
	 return true;
}


//------------------------------------------------------------------------

bool phantom_hl60::draw(i3d::Image3d<i3d::GRAY16>& img_phantoms, 
								i3d::Image3d<LABEL_TYPE>& img_labels)
{
	 if (!_create_shape(img_labels))
		  return false; 
	 
	 _create_texture(img_labels, img_phantoms); 
	 return true;
}


//------------------------------------------------------------------------
template <class T, class L> void 
phantom_hl60::_create_texture(const i3d::Image3d<L> &img_labels,
										i3d::Image3d<T> &img_phantoms)
{
	 T condensation = (T) (std::numeric_limits < T >::max () * 
							(1.0f - GenFromRange(0.0f, condensation_var/100.0f)));

	 L lbl = img_labels.GetMaxValue();

	 if (lbl == (L) 0) // there is no object, do nothing
		  return;

	 echo("lbl: " << (int) lbl);

	 Image3d<bool> mask;
	 Image3d<T> texture;

	 GetLabelMask(img_labels, lbl, mask);
	 texture.CopyMetaData(mask);

#ifdef DEBUG
	 mask.SaveImage("debug-hl60-mask.ics");
#endif
	
	 for (size_t i=0; i<mask.GetImageSize(); i++)
	 {
		  if (mask.GetVoxel(i) == true)
				texture.SetVoxel(i, condensation);
	 }

	 echo("labels size: " << img_labels.GetSize());
	 echo("phantoms size: " << img_phantoms.GetSize());
	 echo("mask size: " << texture.GetSize());
 
    // crop the nucleoli
    echo ("cropping nucleoli");
    _crop_nucleoli (texture, 
						mask, 
						num_of_nucleoli, 
						size_of_nucleoli,
						condensation);
	 echo("cropping finished");

    // add Perlin noise
    float blob_size = GenNum (blob_size_ref, blob_size_var);
    float smoothness = GenNum (smoothness_ref, smoothness_var);

    echo ("adding Perlin noise");

    AddPerlin (texture, blob_size, smoothness, smoothness, contrast);
	 echo("Perlin noise added");

#ifdef DEBUG
    texture.SaveImage("debug-hl60-texture-cropped.ics");
#endif

    // copy the new object into the image with phantoms
	 StickToImage(img_phantoms, texture, mask);
}

//------------------------------------------------------------------------
template <class L>
bool phantom_hl60::_create_shape(i3d::Image3d<L>& img_labels)
{
	 // prepare the initial estimate of object (hl60) dimensions
	 Vector3d<float> r;
	 r.x = GenNum (radius.x, (float) r_variance);

	 // This line prevents the cell to be too small or too large.
	 // If the cell is elongated along x-axis, this line causes
	 // shortening in y-axis, and vice versa.
	 r.y = radius.y * (1.0f - ((r.x-radius.x)/radius.x));
	 //r.y = GenNum (radius.y, (float) r_variance);
	 r.z = GenNum (radius.z, (float) r_variance/2.0f);

	 // define object rotation
	 float xy_rotation = 
				0.9f*GenFromRange ((float)-rot_freedom, (float)rot_freedom),

			 xz_rotation = 
				0.9f*GenFromRange ((float)-rot_freedom, (float)rot_freedom);

	 // create the image containing the initial (rough) object mask
	 Image3d<bool> initial;
	 Vector3d<size_t> sz;
	 float max_r = std::max(r.x, std::max(r.y, r.z));
	 Vector3d<float> centre(max_r), dims = centre * 2.0f;

	 sz = MicronsToPixels(dims, img_labels.GetResolution());
	 initial.MakeRoom(sz + 2); // plus some boundary
	 initial.SetResolution(img_labels.GetResolution());
	 initial.Clear();

    // create the initial object
    AddEllipsoid (initial, 
						centre.x, centre.y, centre.z, 
						r.x, r.y, r.z, 
						xy_rotation, xz_rotation);

	 /*
#ifdef DEBUG
	 initial.SaveImage("debug-init.ics");
#endif
*/

    if (initial.GetMaxValue() == false)
    {
			 throw ERROR_DETAILS("The object you want you generate is smaller" \
										" than one voxel.\n" \
										" Solution: Increase the resolution.");
    }


	 // find the suitable position (if needed) and put the initial object
	 // into dedicated position in an accumulator image
	 Image3d<bool> accumulator;
	 accumulator.CopyMetaData(img_labels);
	 accumulator.Clear();

    // find the available position if not given and put the
    // initial object into the accumulator image
	 if (position == NULL)
	 {
		  Image3d<bool> occupied;
        Threshold(img_labels, occupied, L(1), std::numeric_limits<L>::max());

        Vector3d<float> *random_pos = 
					 find_position_in_cluster(initial, 
													  occupied, 
													  cluster_effect, 
													  influence_of_gravity);

		  if (random_pos == NULL)
				return false;

		  MergeBinaryImages(accumulator, initial, *random_pos - centre);

		  delete random_pos;
	 }
	 else
	 {
		  MergeBinaryImages(accumulator, initial, *position - centre);
	 }

	 /*
#ifdef DEBUG
	 accumulator.SaveImage("debug-accumulator.ics");
#endif
*/

	 echo("deformation started");

    // read the parameters which drive the deformation
	 float deformation_level = GenNum (ishape, ishape_var);

	 Image3d<float> img_speed;
	 img_speed.CopyMetaData(img_labels);
	 GenerateSpeedImage(img_speed, (r.x+r.y+r.z)/4.5f);

	 Image3d<bool> free_space;
	 Threshold(img_labels, free_space, L(0), L(0));

	 // draw a border line in the free space to prevent the possible object
	 // to touch the border
	 for (size_t i=0; i<free_space.GetImageSize(); i++)
	 {
		  if (free_space.OnBorder(i))
		  {
				free_space.SetVoxel(i, false);
		  }
	 }

	 // perform the deformation of an initial object (inserted into
	 // the accumulator image - this insertion guarantees that the neighbouring
	 // object won't touch or overlap each other during the deformation)
#ifdef DEBUG
	 accumulator.SaveImage("debug-hl60-nondeformed.ics");
#endif

    Deformation (accumulator, 
					  &free_space,
					  img_speed, 
					  deformation_level);

#ifdef DEBUG
	 accumulator.SaveImage("debug-hl60-deformed.ics");
#endif

	 echo("deformation finished");

    // add the mask to the labeled image to store the information concerning
    // this new mask
    return AddNewLabel(img_labels, accumulator);
}

//------------------------------------------------------------------------
// internal function - generates nucleoli in the cell nucleus preserving
// the area dedicated by the nucleus mask
//------------------------------------------------------------------------
void phantom_hl60::_gen_nucleoli(const Image3d<bool> &nucleus_mask,
										 Image3d<bool> &nucleoli_mask,
										 int count,
										 int size)
{
	 nucleoli_mask.CopyMetaData(nucleus_mask);
	 nucleoli_mask.Clear();

	 Image3d<bool> occupied = nucleus_mask;
	 Negative(occupied);

	echo("\t number of nucleoli: " << count);

	// loop for each nucleolus
	for (int i=0; i<count; i++)
	{
		 Vector3d<float> r;
		 r.x = GenNum (radius.x, (float) r_variance) * (size / 100.0f);
		 r.y = GenNum (radius.y, (float) r_variance) * (size / 100.0f);
		 r.z = GenNum (radius.z, (float) r_variance) * (size / 100.0f);
		 float max_r = std::max(r.x, std::max(r.y, r.z));
		 Vector3d<float> centre(max_r), dims = centre * 2.0f;
		 Vector3d<size_t> sz = MicronsToPixels(dims, 
															nucleus_mask.GetResolution());

		 Image3d<bool> initial;
		 initial.MakeRoom(sz);
		 initial.SetResolution(nucleus_mask.GetResolution());
		 initial.Clear(); 

	 	 float xy_rotation = 0.9f*GenFromRange ((float)-rot_freedom, (float)rot_freedom),
				 xz_rotation = 0.9f*GenFromRange ((float)-rot_freedom, (float)rot_freedom);

		 // generate the initial nucleolus shape
		 AddEllipsoid(initial, 
						  centre.x, centre.y, centre.z,
						  r.x, r.y, r.z,
						  xy_rotation, xz_rotation);
		
		 Vector3d<float> *pos;
		 pos = find_position_uniformly(initial, occupied, 0.0f);
		 if (pos == NULL)
		 {
	        throw ERROR_DETAILS("Cannot find a suitable position for " \
										 "a new nucleolus.\n" \
										 "Try to enlarge your VOI.");
		 }

		 // insert the nucleolus to the following images
		 // - occupied (it guarantees that the new object won't overlap)
		 // - nucleoli_mask (to save the results)
		 MergeBinaryImages(occupied, initial, *pos - centre);
		 MergeBinaryImages(nucleoli_mask, initial, *pos - centre);

		 delete pos;
	}
}

//--------------------------------------------------------------------------
// Function _crop_nucleoli cuts out selected number of nucleoli from the given 
// image. The nucleoli are then deformed and put into the image of the 
// original nucleus.
//--------------------------------------------------------------------------
template <class T> 
void phantom_hl60::_crop_nucleoli(Image3d<T> &texture,
											 const Image3d<bool> &nucleus_mask,
											 int count, 
											 int size,
											 T value) 
{
	 echo("\t count: " << count);
	 echo("\t size: " << size << "% of nucleus size ");

	if (count == 0)
		 return;

	// gen nucleoli
	Image3d<bool> nucleoli_mask; 
	_gen_nucleoli(nucleus_mask, nucleoli_mask, count, size);

	// deform nucleoli
	Image3d<float> img_speed;
	img_speed.CopyMetaData(nucleus_mask);
	GenerateSpeedImage(img_speed, 3.0); // TODO: tady je to abs velicina!
	float deformation_level = GenNum (ishape, ishape_var);
	Deformation(nucleoli_mask, &nucleus_mask, img_speed, deformation_level);

#ifdef DEBUG
	nucleoli_mask.SaveImage("debug-hl60-nucleoli.ics");
#endif

	// the nucleoli content intensity is approx 20% of the surrounding
	// environment 
	T level = (T)(0.20 * value); // TODO: this value should be in INI file too

	for (size_t i=0; i<texture.GetImageSize(); i++)
	{
		if (nucleoli_mask.GetVoxel(i) == true)
		{
			texture.SetVoxel(i, level);
		}
	}
}

//------------------------------------------------------------------------

extern "C" DLL_EXPORT phantom_hl60* create()
{
	 return new phantom_hl60;
}

//------------------------------------------------------------------------

extern "C" DLL_EXPORT void destroy(phantom_hl60* ph)
{
	 delete ph;
}
//------------------------------------------------------------------------
