/*---------------------------------------------------------------------

hl60.h

This file is part of CytoPacq

Copyright (C) 2007-2013 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Author: David Svoboda

Description: A plugin creating the digital phantom of HL60 nucleus.

-------------------------------------------------------------------------*/

#ifndef _PHANTOM_HL60_
#define _PHANTOM_HL60_

#include "generic.h"

//------------------------------------------------------------------------

class phantom_hl60 : public phantom_with_position
{ 
  private:			 
			 // ---------- internal variables -----------------
			 /// 3D size of the ellipsoid - the initial rough object (microns)
			 t_triplet radius; 
			 /** How much the axes can vary. The number defines the fuzziness 
			  * of the previous property. It is given in percents and defines 
			  * the range [radius]*<1-r_var/100;1+r_var/100> within which 
			  * the new value should be randomly found. Note that all the variance-like 
			  * properties will be defined in the same manner (percents).
			  */
			 int r_variance; 
			 /// Fuzziness of the object canonical (aligned with axis) position (percents). 
			 int rot_freedom; 
			 /** How much the quality of dying varies within the image. The nucleus
			  * with more condensed chromatin should be brighter while the others should
			  * be darker. The value is defined as a fraction of the intensity 
			  * of the brightest nucleus (percents).
			  */
			 float condensation_var; 
			 /** The fuzziness of the ellipsoid shape. This parameter drives 
			  * the level of generated ellipsoid deformation. It is a real number 
			  * in the range <0;oo>. Zero means no deformation. As the deformation 
			  * function is very sensitive small numbers are good enough to get 
			  * sufficiently deformed objects.
			  */
			 float ishape;
			 /** If we want to avoid the similarity of the deformation 
			  * within a number of object some further fuzziness driving 
			  * the deformation level is needed as well (percents).
			  */
			 float ishape_var; 
			 /** The object content contains dark areas - nucleoli. Those are 
			  * the nucleus components not affected by a dye and hence not visible (number)
			  * Implicit value is '1' or '2'
			  */
			 int num_of_nucleoli; 
			 /// Size of the nucleoli as a fraction of the nucleus size (percents). 
			 int size_of_nucleoli; 
			 /**  The nucleus content is defined using a texture called Perlin noise. 
			  * The blob size (in microns) defines the size of blobs appearing 
			  * in the texture.
			  */
			 float blob_size_ref; 
			 /// Fuzziness of the blob size (percens).
			 float blob_size_var; 
			 /** The Perlin noise may be smooth as well as containing coarse grains. 
			  * This property drives the smoothness of the texture. The values are within 
			  * the range <0;oo>. Lower value means lower level of smoothness (number)
			  */
			 float smoothness_ref; 
			 /// The fuzziness of the smoothness (percents).
			 float smoothness_var; 
			 /** The texture histogram may be very wide (high contrast) or 
			  * narrow (low contrast); (number <0;oo)
			  */
			 float contrast; 
	
			 // ---------- internal methods --------------------
			 template <class L>
			 bool _create_shape(i3d::Image3d<L> &);

			 template <class T, class L>
			 void _create_texture(const i3d::Image3d<L> &,
										 i3d::Image3d<T> &);

			 template <class T>
			 void _crop_nucleoli(i3d::Image3d<T> &, 
										const i3d::Image3d<bool> &,
										int count,
										int size,
										T value); 

			 void _gen_nucleoli(const i3d::Image3d<bool> &, 
									  i3d::Image3d<bool> &, 
									  int count,
									  int size);

  public:                               
			 // -------------- public methods -------------------------
			 virtual void read_params(const t_item_seq &,
											  const t_item_seq &);

			 virtual bool draw(i3d::Image3d<i3d::GRAY8>&, 
									 i3d::Image3d<LABEL_TYPE>&);

			 virtual bool draw(i3d::Image3d<i3d::GRAY16>&, 
									 i3d::Image3d<LABEL_TYPE>&);

			 /// virtual destructor
			 virtual ~phantom_hl60() {};
};

//------------------------------------------------------------------------

#endif
