/*---------------------------------------------------------------------

microsphere.cc

This file is part of CytoPacq

Copyright (C) 2007-2013 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Author: David Svoboda

Description: A plugin creating the digital phantom of microsphere.

-------------------------------------------------------------------------*/

#include <i3d/threshold.h>

#include "sim_macros.h"
#include "microsphere.h"

#include "src/collisions.h"
#include "src/add_shape.h"

using namespace i3d;

//------------------------------------------------------------------------
extern "C" DLL_EXPORT const char *get_id()
{
	 return "Microsphere";
}

//------------------------------------------------------------------------
void phantom_microsphere::read_params(const t_item_seq &phantom_properties,
												  const t_item_seq &common_properties)
{
	 phantom_with_position::read_params(phantom_properties,
													common_properties);

	 diameter = phantom_properties["diameter"]; 
}

//------------------------------------------------------------------------

bool phantom_microsphere::draw(i3d::Image3d<i3d::GRAY8>& img_phantoms, 
										 i3d::Image3d<LABEL_TYPE>& img_labels)
{
	 if (!_create_shape(img_labels))
		  return false; 
	 
	 _create_texture(img_labels, img_phantoms); 
	 return true;
}

//------------------------------------------------------------------------

bool phantom_microsphere::draw(i3d::Image3d<i3d::GRAY16>& img_phantoms, 
										 i3d::Image3d<LABEL_TYPE>& img_labels)
{
	 if (!_create_shape(img_labels))
		  return false; 
	 
	 _create_texture(img_labels, img_phantoms); 
	 return true;
}


//------------------------------------------------------------------------
template <class L>
bool phantom_microsphere::_create_shape(i3d::Image3d<L> &img_labels)
{
  // establish the initial (small) image
  Image3d<bool> initial;
  Vector3d<size_t> sz;
  Offset r_offset;

  sz = MicronsToPixels(Vector3d<float>(diameter), 
							  img_labels.GetResolution());
  initial.MakeRoom(sz + 2); // plus some boundary
  initial.SetResolution(img_labels.GetResolution());
  initial.Clear();

  // draw a sphere with the given radius to the image centre
  Vector3d<float> centre(diameter/2);
  AddEllipsoid (initial, 
					 centre.x, 
					 centre.y, 
					 centre.z, 
					 diameter/2, 
					 diameter/2, 
					 diameter/2);

  echo("microsphere generated");

  if (!initial.GetVoxelData().max())
  {
			 throw ERROR_DETAILS("The object you want you generate "\
										"is smaller than one voxel.\n" \
					" Solution: Increase the resolution. If it is not possible," \
					" change the object size.");
  }

  // find the available position if not given and put the
  // initial object into the labeled image
  if (position == NULL)
  {
		// try to find suitable position
		Image3d<bool> occupied;
		Threshold(img_labels, occupied, L(1), std::numeric_limits<L>::max()); 
	
		echo("calling 'find_position' routine");
		Vector3d<float> *random_pos = 
				  find_position_in_cluster(initial, 
													occupied, 
													cluster_effect, 
													influence_of_gravity);

		if (random_pos == NULL)
			 return false;

		echo("*** random_pos: " << *random_pos);
		echo("*** centre: " << centre);
		echo("*** image offset: " << img_labels.GetOffset());

		r_offset = (*random_pos - centre) + img_labels.GetOffset();

		delete random_pos;
  }
  else
  { 
		r_offset = (*position - centre) + img_labels.GetOffset();
  }

  initial.SetOffset(r_offset);

  return AddNewLabel(img_labels, initial);
}

//------------------------------------------------------------------------
template <class T, class L> 
void phantom_microsphere::_create_texture(
	const i3d::Image3d<L> &img_labels,
	i3d::Image3d<T> &img_phantoms)
{
	 // TODO: sel by uz pridat photobleachning - uz tady mame 
	 // dopropagovany potrebny parametr 

  // get mask image
  Image3d<bool> mask;
  L max_label = img_labels.GetMaxValue();
  GetLabelMask(img_labels, max_label, mask);

  // prepare texture image
  Image3d<T> texture;
  texture.CopyMetaData(mask);

  for (size_t i=0; i<mask.GetImageSize(); i++)
  {
		if (mask.GetVoxel(i))
			 texture.SetVoxel(i, std::numeric_limits<T>::max());
  }

  // put the modified image back
  StickToImage(img_phantoms, texture, mask);

//  echo("MAXIMUM:" << (size_t) texture.GetVoxelData().max());
}

//------------------------------------------------------------------------

extern "C" DLL_EXPORT phantom_microsphere* create()
{
	 phantom_microsphere* ph = new phantom_microsphere;
	 return ph;
}

//------------------------------------------------------------------------

extern "C" DLL_EXPORT void destroy(phantom_microsphere* ph)
{
	 delete ph;
}
//------------------------------------------------------------------------
