/*---------------------------------------------------------------------

diagram.h

This file is part of CytoPacq

Copyright (C) 2009-2013 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Authors: 
- David Svoboda <svoboda@fi.muni.cz> May 2010
- Ondrej Homola <ondra.homola@gmail.com> June 2009

Description: The implementation of some geometric primitives in 3D space
             (plane, polyhedron, voronoi diagram). 

-------------------------------------------------------------------------*/

#ifndef __VORONOI_DIAGRAM__
#define __VORONOI_DIAGRAM__

#include <vector>
#include <i3d/vector3d.h>

//--------------------------------------------------------------------------

/// Index (data-type)
typedef size_t INDEX;

/// Floating point (data-type )
typedef double REAL;

/// Point in 3D space (data-type)
typedef i3d::Vector3d<REAL> Position;

/// Vector in 3D space (data-type)
typedef i3d::Vector3d<REAL> Vector;

//--------------------------------------------------------------------------
/**
 * Class Plane
 *
 * This class defines an oriented plane in 3D space.
 *
 * The mathematical definition of a plane in 3D is following
 *
 *  Ax + By + Cz + D = 0
 *
 *  (A,B,C) ... normal vector
 *  D ... offset
 *
 */
//--------------------------------------------------------------------------
class Plane
{
  protected:
			 /// Normal vector of a plane
			 Vector normal_vector;
			 /// Plane offset (according to the definition of a plane)
			 REAL offset;

  public:
			 /// Constructor accepting three 3D points
			 Plane(const Position &, const Position &, const Position &);
			 /// Constructor accepting normal vector and offset
			 Plane(const Vector &, REAL);

			 /**
			  * Tests whether the given point lies on the left or right side
			  * of this plane.
			  */
			 bool has_on_negative_side(const Position &_point) const;

			 /**
			  * Tests whether the given point lies on the left or right side
			  * of this plane.
			  */
			 bool has_on_positive_side(const Position &_point) const;

			 /// Change the orientation of a plane (normal vector is inverted).
			 Plane opposite() const;
};

//--------------------------------------------------------------------------
/**
 * Class Polyhedron 
 *
 * The Polyhedron class that defines one Voronoi region in 
 * the 3D Voronoi diagram.
 *
 * The polyhedron is defined via planes (interection of halfspaces defined
 * using these plane) + the list of all vertices is also stored.
 *
 */
//--------------------------------------------------------------------------
class Polyhedron
{
public:
	/** 
	 * Constructor for naive representation of a convex polyhedron.  
	 */
	Polyhedron(const std::vector<Plane> &, const std::vector<Position> &);

	/**
	 * Implicit constructor 
	 */
	Polyhedron();

	/** 
	 * Returns a point of this object's bounding box, 
	 * with minimal coordinates (top-left-front).
	 */
	Position GetTopBoundingPoint() const;

	/** 
	 * Returns a point of this object's bounding box, 
	 * with maximal coordinates (lower-right-back).
	 */
	Position GetLowBoundingPoint() const;

	/// This method gives the number of planes that belong to this polyhedron.
	size_t GetNumberOfPlanes() const { return planes.size(); }

	/// This method gives the number of vertices that belong to this polyhedron.
	size_t GetNumberOfVertices() const { return vertices.size(); }

	/** 
	 * Establish bounding box of this polyhedron.
	 * This method helps when checking if an inspected point lies inside
	 * or outside the polyhedron. The searched space is significantly reduced.
	 */
	void FindBoundingPoints();

	/** 
	 * Get the centre of the polyhedron.
	 * As the class Polyhedron is utilized as a representation of Voronoi
	 * regions, the center is never evaluated as a weighted sum of all the
	 * interior voxels. The center is a point that exactly matches the 
	 * one triangulated point in Delaunay triangulation
	 */
	Position GetCenter() const { return center; }

	/// Define the center of polyhedron
	void SetCenter(const Position &_center) { center = _center; };
	
	/** 
	 * Method for rasterization. Returns true if and only 
	 * if the given point is located inside this Polyhedron.
	 */
	bool contains(const Position& ) const;

	/// Add a new plane to the polyhedron.
	void AddPlane(const Plane& _plane);

	/// Add a new vertex to the polyhedron.
	void AddVertex(const Position& _vertex);

	/// Increment the volume of polyhedron by the given number
	void IncrementVolume(double amount);

	/// Get the volume of polyhedron
	double GetVolume() const { return volume; };

private:
	/// A flag that notifies whether the bounding box is defined or not.
	bool bounding_points_found;

	/// Volume of the polyhedron
	double volume;

	/// List of planes that define the polyhedron
	std::vector<Plane> planes;

	/// List of vertices that define the polyhedron
	std::vector<Position> vertices;

	/// Center of polyhedron
	Position center;

	/// Top part of bounding box
	Position top_bounding_point;

	/// Bottom part of bounding box
	Position low_bounding_point;
}; 


//--------------------------------------------------------------------------
/**
 * Class VoronoiDiagram
 *
 * This class represents a structure of Polyhedron objects. Its methods
 * implement the creation of 3D VoronoiDiagram and manipulation with it.
 *
 * For the creation of Voronoi diagram 
 * the QHULL program (http://www.qhull.org/) is utilized.
 */
//--------------------------------------------------------------------------

class VoronoiDiagram 
{
public:

   // constructor
	VoronoiDiagram();

	// destructor
	~VoronoiDiagram();

	/**
	This method converts the sets of 3d points into a voronoi diagram.
	/param cell_points vector of cell points.
	/param control_points vector of control points.
	*/
	void Create(const std::vector<Position> &cell_points, 
					const std::vector<Position> &control_points);

	/// This method returns the list of voronoi regions
	const std::vector<Polyhedron>& GetRegions();

	/// This method return the list of skewed (deformed) voronoi regions 
   const std::vector<Polyhedron>& GetSkewedRegions(double, double, double);


private:

	/**
	 * QHULL program requires some input/output file buffer. These
	 * are the name of these buffer. The names are created when 
	 * the instance of this class is created. The files are created 
	 * when calling QHULL program. Finally, the files bearing these names 
	 * are deleted in destructor.
	 */
	std::string qhull_input_fname, qhull_output_fname, qhull_log_fname;

	/// List of voronoi region
   std::vector<Polyhedron> regions;

	/** 
	 * This method is responsible for executing the QHULL program.
	 */
	void CallQhull();

	/// This method clears the internal variables.
	void Clear();

	/**
	 * This method is responsible for parsing this QHULL output and converting
	 * it into internal structures + skewing the whole voronoi diagram.
	 */
	void ParseQhullOutput(double, double, double);

	/** 
	 * This method is responsible for preparing the input file buffer
	 * for QHULL program.
	 */
	void PrepareQhullInput(const std::vector<Position>&,
								  const std::vector<Position>&);
};

//--------------------------------------------------------------------------


#endif  

