/*---------------------------------------------------------------------

histogram.cc

This file is part of CytoPacq

Copyright (C) 2007-2013 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Author: David Svoboda

Description: The input is the list of images. The output is the intensity histogram including mean and variance.

-------------------------------------------------------------------------*/

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <iomanip>

#include "i3d/image3d.h"
#include "i3d/imgfiles.h"
#include "i3d/transform.h"
#include "i3d/toolbox.h"
#include "i3d/histogram.h"
#include "i3d/threshold.h"
#include "i3d/morphology.h"
#include "i3d/regions.h"

#define sqr(x) ((x)*(x))

using namespace std;
using namespace i3d;

typedef GRAY8 TYP;

const size_t gmax = std::numeric_limits<TYP>::max();

int main(int argc, char* argv[]) 
{
	if (argc < 2)
		{
			cerr << "Syntax: " << endl;
			cerr << argv[0] << " <list of images> " << endl;
			cerr << "    <list of images> ... a file containing the list of images" << endl;
			cerr << endl;
			exit(-1);
		}

	ifstream f(argv[1]);
	Image3d<TYP> img;
	Histogram hist(gmax);

	Image3d<TYP> crop;
	//VOI<size_t> voi(35,35,10,120,110,65);
	VOI<PIXELS> voi(25,25,0,130,130,65);

	vector<float> sum(gmax, 0.0f), sum_sq(gmax, 0.0f);

	string img_name;
	size_t counter = 0;

	// loop for all the images
	while (!f.eof())
	{
		f >> img_name;

		cout << "processing: " << img_name << endl;

		img.ReadImage(img_name.c_str());
		crop.CopyFromVOI(img, voi);
		img = crop;

		// clear histogram before each loop
		for (size_t i=0; i<gmax; i++)
		{
			hist[i] = 0;
		}
		IntensityHist(img, hist);
		size_t hmax = hist.max();

		for (size_t i=0; i<gmax; i++)
		{
			// we will normalize histogram value to eliminate troubles with
			// images of different sizes
			float value;
			if (hist[i] == 0)
				 value = 0;
			else
				 value = log(((float)hist[i]))/log((float)img.GetImageSize());
				 //value = log(((float)hist[i]))/log(hmax);

			//float value = ((float)hist[i])/img.GetImageSize();
			//float value = ((float)hist[i])/hmax;
			sum[i] += value;
			sum_sq[i] += sqr(value);
		}

		counter++;
	}

	cout << "Mean histogram and its variance" << endl;
	cout << "- number of images: " << counter << endl;

	const char *output_name = "hist.txt";
	ofstream fout(output_name);

	for (size_t i=0; i<gmax; i++)
	{
		float mean = sum[i]/counter;
		float var = sqrt((float)sum_sq[i]/counter - sqr(mean));

		cout << i << "   " << mean << "   " << var << endl;
		fout << i << "   " << mean << "   " << var << endl;
	}

	cout << "The results stored in: " << output_name << endl;
  
  return 0;
}
