/*---------------------------------------------------------------------

make_preview.cc

This file is part of CytoPacq

Copyright (C) 2007-2013 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Author: David Svoboda

Description: This program crops out the slice of the given order
from the supplied image.

-------------------------------------------------------------------------*/

#include "i3d/image3d.h"
#include "i3d/threshold.h"
#include "i3d/histogram.h"
#include "i3d/transform.h"
#include <stdio.h>
#include <iostream>
#include <vector>
#include <string>
#include <locale.h>

using namespace std;
using namespace i3d;

template <class T> 
void CutSlice(const char *in, double z, const char *out)
{
	 Image3d<T> img_in(in), img_out;
	 unsigned z_pos;

	 if ((z < 1) && (z >= 0))
	 {
		  z_pos = (unsigned) floor(z * img_in.GetSizeZ());
	 }
	 else
	 {
		  z_pos = img_in.GetSizeZ() / 2;
	 }

	 Vector3d<size_t> v = img_in.GetSize();

	 VOI<PIXELS> voi(0, 0, z_pos, v.x, v.y, 1);

	 // copy the slice
	 img_out.CopyFromVOI(img_in, voi);

	 Image3d<GRAY8> img_jpeg;
	 img_jpeg.MakeRoom(img_out.GetSize());

	 // apply linear stretch
	 float maxT = img_out.GetVoxelData().max();
	 float minT = img_out.GetVoxelData().min();
	 float max8 = std::numeric_limits<GRAY8>::max();

	 for (size_t i=0; i<img_out.GetImageSize(); i++)
	 {
		  T value = img_out.GetVoxel(i);
		  GRAY8 new_value = (GRAY8) floor((value-minT)/(maxT-minT)*max8); 

		  img_jpeg.SetVoxel(i, new_value);
	 }

	 // save the preview image
	 img_jpeg.SaveImage(out);
}

int
main (int argc, char *argv[])
{
	 if (argc != 4)
	 {
		  cerr << "Bad input!" << endl;
		  cerr << "syntax: " << argv[0] <<
		 	 "<input image> <slice number> <preview name>" << endl << endl;
		  exit(-1);
	 }

  try
  {
    ImgVoxelType vt = ReadImageType (argv[1]);

    switch (vt)
      {
	  case Gray8Voxel:
				 {
					  CutSlice<GRAY8>(argv[1], atof(argv[2]), argv[3]);
						break;
				 }
	  case Gray16Voxel:
				 {
					  CutSlice<GRAY16>(argv[1], atof(argv[2]), argv[3]);
						break;
				 }
	  case FloatVoxel:
				 {
					  CutSlice<float>(argv[1], atof(argv[2]), argv[3]);
						break;
				 }
      default:
        cerr << "Wrong type." << endl;
      }
  }
  catch (IOException & e)
  {
    cerr << e << endl;
    exit (1);
  }
  catch (InternalException & e)
  {
    cerr << e << endl;
    exit (1);
  }
  catch (bad_alloc &)
  {
    cerr << "Not enough memory." << endl;
    exit (1);
  }
  catch (...)
  {
    cerr << "Unknown exception." << endl;
    exit (1);
  }

  return 0;
}
